# -*- coding: utf-8 -*-
"""
Created on Fri Dec  9 15:21:00 2016

@author: akidribm
"""

from sortal import (transformSet, similarityMap, rule, primitiveSort, point2D, coordinate, lineSegment2D, line2D, label, weight, disjunctiveSort, attributeSort, affineTransform,  vector2D, vector2Dfloat, planeSegment2D, point3D, vector3D, lineSegment3D, attributeSort, disjunctiveSort, weight, coForm, label, transformSet)
from sort_types import sort_types
from sympy import pi, sin, cos
    
"""
=============
FORM CREATION
=============
'Create' methods generate and return forms based on input. As part of the API, they are imported by 
importing create from the api_create file, i.e. from api_create import create.

All sort types are initialized and stored in the file sort_types.py, inside the dictionary sort_types.
This is imported via 'from sort_types import sort_types'

Vectors are created by calling either vector2D or vector3D, as follows:
v1 = vector2D(x,y) where (x,y) are coordinates
v2 = vector3D(x,y,z) where (x,y,z) are coordinates

=======
METHODS
=======
    
Create.createForm where a form of individuals is created from a list of individuals

                  EXAMPLE: a form of line segments where ls1 = lineSegment2D(sort_types['lineSeg2D'], start, end) and so on
                  new_form1_list = [ls1, ls2, ls3, ls4, ls5, ls6, ls7, ls8, ls9, ls10] 
                  new_form1 = create.createForm(sort_types['lineSeg2D'], new_form1_list) 
                   
                  OR a form of points where p1 = point2D(sort_types['pointSort2D'], x, y) and so on
                   
                  new_form1_list = [p1, p2, p3, p4, p5, p6, p7, p8, p9, p10] 
                  new_form1 = create.createForm(sort_types['lineSeg2D'], new_form1_list)
                   
Create.lineSegForm where a form of line segments is created from an input of vectors as lists within a list
                   Each vector pair corresponds to a line segment, where index 0 of the pair is the start
                   of the line segment and index 1 of the pair is the end of the line segment
                  
                   This method will raise an error if the sort type passed is not based on 
                   lineSegment2D or lineSegment3D, i.e. sort_types['lineSeg2D'] OR sort_types['lineSeg3D']
                     
                   new_form2_set = [[vb1, vb2], [vb2, vb3], [vb3, vb4], [vb4, vb1]]
                   new_form2_set = create.lineSegForm(sort_types['lineSeg2D'], form2_set)
                  
                   OR if with 3D vectors (vector3D), 
                  
                   new_form2_set = create.lineSegForm(sort_types['lineSeg3D'], form2_set)
                  
Create.polygonForm where a form containing a polygon is created from an input of a list of vectors
                   These vectors serve as vertices and need to be listed in the order they will be connected
                   It is assumed that the last vector will be connected to the first vector
                   In the example, v10 will be connected to v1 to close the polygon
                    
                   This method will raise an error if the sort type passed is not based on 
                   lineSegment2D or lineSegment3D, i.e. sort_types['lineSeg2D'] OR sort_types['lineSeg3D']
                    
                   new_form3_set = [v1, v2, v3, v4, v5, v6, v7, v8, v9, v10]
                   new_form3 = create.polygonForm(sort_types['lineSeg2D'], new_form3_set)
                    
                   OR if with 3D vectors (vector3D), 
                  
                   new_form2_set = create.polygonForm(sort_types['lineSeg3D'], new_form3_set)
                  
Create.planeForm where a form bounded by the elements in the bound argument with the sort_type (planeSort) is created. 
                 Attributes can be added later on if the planeSort2D defined in sort_types is based on attributeSort,
                 e.g. planeSort2D = attributeSort('planeAttrib2D', planeS2D, weightSort) 
                 OR planeSort2D = attributeSort('planeAttrib2D', planeS2D, labelSort) 
                 
                 This method will raise an error if the sort type passed is not based on planeSegment2D.
                
                 bound = [v1, v2, v3, v4]
                 formP = create.planeForm(sort_types['planeSort2D'], bound)
                 In this case planeSort is an attribute sort with a weight attribute (as defined inside the file sort_types)
                 
                 To add an attribute like weight to the first/nth plane inside formP:
                     
                 Create the weight attribute first: weight = weight(sort_types['weight'], 255)
                 
                 The plane/s inside formP are housed inside formP.elements
                 EXAMPLE: To add an attribute to the first plane inside form: formP.elements[0].addAttribute(weight)

Create.pointForm where a form of points is created from the inputs sort_type, and the (x,y) or (x,y,z) coordinates of an 
                 initial point which is inserted in the form. New points can be created and inserted later 
                 on into the form. This method will raise an error if the sort type passed is not based on point2D or point3D.

                 formPt = create.pointForm (sort_types['pointSort2D'], 1, 1) OR formPt = create.pointForm (sort_types['pointSort3D'], 1, 1, 1)
                 
                 To insert another point,
                 point = point2D(sort_types['pointSort2D'], 1, 0) OR point = point3D(sort_types['pointSort3D'], 1, 0, 1)
                 formPt.insert(point)

Create.addAttributeAll takes a form and an attribute as inputs. It adds the attribute (weight/label) to 
                       all points inside the form.
                       
                       This method will raise an error if the sort type of the form is not based on attributeSort.
                       Note: A check for the attribute sort type will be added later on.
                    
                       1. Create label attribute
                       label = label(sort_types['label'], 'black')
                    
                       2. Initially add one point at (1,1) to the form --> form has 1 point: (1,1)
                       formPt = create.pointForm (sort_types['pointSort2D'], 1, 1) 
                       
                       Transform point inside point form by angle a
                       formPt1 = transform(formPt1, a) 

                       3. Add the other point at (1,1) to the form --> form has 2 points: (sqrt(3)/3,1) and (1,1)
                       point2 = point2D(sort_types['pointSort2D'], 1, 1)   
                       formPt.insert(point2) 
                    
                       4. Add black label to both points
                       formPt = create.addAttributeAll(formPt, label) 

Create.metaForm takes a list of forms as input and inserts them into a disjunctive Sort metaform. This method will raise an
                error if the sort type passed is not based on disjunctiveSort.

                forms = [formP1, formPt1]
                formM1 = create.metaForm(sort_types['disjunctive2D'], forms)
"""

class create:
    
    def createForm(self, sort_type, sort_set):
        form = sort_type.newForm()
        for i in sort_set:
            form.insert(i)
        return form

    def lineSegForm(self, sort_type, vector_set):
        if sort_type.baseSort().characteristic != lineSegment2D and sort_type.baseSort().characteristic != lineSegment3D:
            print ("ERROR: Only sort types based on lineSegment2D or lineSegment3D can be passed to create.lineSeg_form.")
            raise TypeError
        form = sort_type.newForm()
        for i in vector_set:
            if sort_type.baseSort().characteristic == lineSegment2D:
                line = lineSegment2D(sort_type, i[0], i[1])
            elif sort_type.baseSort().characteristic == lineSegment3D:
                line = lineSegment3D(sort_type, i[0], i[1])
            form.insert(line)
        return form
    
    def polygonForm(self, sort_type, vector_set):
        if sort_type.baseSort().characteristic != lineSegment2D and sort_type.baseSort().characteristic != lineSegment3D:
            print ("ERROR: Only sort types based on lineSegment2D or lineSegment3D can be passed to create.polygon_form.")
            raise TypeError
        i = 0
        form = sort_type.newForm()
        vector_set.append(vector_set[0])
        while i < len(vector_set) - 1:
            if sort_type.baseSort().characteristic == lineSegment2D:
                line = lineSegment2D(sort_type, vector_set[i], vector_set[i+1])
            elif sort_type.baseSort().characteristic == lineSegment3D:
                line = lineSegment3D(sort_type, vector_set[i], vector_set[i+1])
            form.insert(line)
            i+=1
        return form
        
    def planeForm(self, sort_type, bound):
        if sort_type.baseSort().characteristic != planeSegment2D:
            print ("ERROR: Only sort types based on planeSegment2D can be passed to create.planeForm.")
            raise TypeError
        plane = planeSegment2D(sort_type, bound) # Initialize the plane segment with bound
        formP = sort_type.newForm() # Creating the coForm for containing the plane segment          
        formP.insert(plane) # Add/insert the plane segment into its form(coform)s
        return formP
        
    def pointForm(self, sort_type, x, y, z = 0):
        if sort_type.baseSort().characteristic != point2D and sort_type.baseSort().characteristic != point3D:
            print ("ERROR: Only sort types based on point2D or point3D can be passed to create.pointForm.")
            raise TypeError
        if sort_type.baseSort().characteristic == point2D:
            point = point2D(sort_types['pointSort2D'], x, y) 
        elif sort_type.baseSort().characteristic == point3D:
            point = point3D(sort_types['pointSort3D'], x, y, z)
        formPt = sort_type.newForm()
        formPt.insert(point)
        return formPt
    
    def addAttributeAll(self, form, attribute):
        if form.sort.__class__ != attributeSort:
            print ("ERROR: Only forms with sort types based on attributeSort can be passed to create.addAttributeAll.\n")
            raise TypeError
        # Later on, to include check for attribute's sort type
        for i in form.elements: 
            i.addAttribute(attribute)
        return form
        
    def metaForm(self, sort_type, forms):
        if sort_type.__class__ != disjunctiveSort:
            print ("ERROR: Only sort types based on disjunctiveSort can be passed to create.metaForm.")
            raise TypeError
        formM = sort_type.newForm()
        for i in forms:
            formM.insert(i)
        return formM