# -*- coding: utf-8 -*-
"""
Created on Tue Aug 23 12:23:02 2016

@author: akibdpt
"""
from .line import line2D, line3D
from copy import deepcopy
from ..struct.coordinate import coordinate
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform2D, affineTransform3D
"""
 * A line-segment is a connected and bounded segment of a line with an
 * interval behavior. The line defines the co-descriptor of the line-segment,
 * the boundary of the segment is defined by the start and end positions of the
 * line-segment.
 * 
 * The LineSegment class extends on the Line class and
 * implements the Interval interface.
 * It defines the characteristic individual for line-segments.
 * A line-segment is represented as a line with two rational
 * scalars specifying the tail and head relative to the line's root.
 * This characteristic individual accepts no parameters.
 * It specifies a {sortal.map.similarityMap} as default.
 * Forms of line-segments adhere to an interval behavior.
"""
class lineSegment2D(line2D):
    """
    initialize a line segment
    initialize a line containing it first
    find the lesser and greater points between the 2 points passed in (start 
    and end) and store them as self.start and self.end respectively
    """
    def __init__(self, sort, start = vector2D(0,0), end = vector2D(0,0)):
        super().__init__(sort, start, end)     
        self.type = 'ind-lineSegment2D'
        if end.type == 'ind-point2D':
            end = end.value
        if start.type == 'ind-point2D':
            start = start.value
        if end.compareValue(start) == -1:
            self.start = end
            self.end = start
        elif end.compareValue(start) == 1:
            self.end = end
            self.start = start
        else:
            self.end = end
            self.start = start

    def __str__(self):
        return self.type+'_'+str(self.start)+','+str(self.end)

    def printVal(self):
        return (str(self.start)+ '|'+ str(self.end)) # + ', direction: ('+str(self.direction.x)+','+str(self.direction.y)+ '), unitDir: ('+str(self.unitDir.x)+','+str(self.unitDir.y)+ '), root: ('+str(self.root.x)+','+str(self.root.y)+ ')'

    def __eq__(self, other):
        if other is None or other.type != 'ind-lineSegment2D' :
            return False
        if self.sort == other.sort:
            return (self.start == other.start) and (self.end == other.end)
        
    def isNil(self):
        if self.end is None or self.start is None:
            return True
        return self.end.x == self.start.x and self.end.y == self.start.y
        
    """
    compare the value of two line segments
    first compare using unit directions
    if unit directions are equal, compare using roots
    if roots are equal, compare the start points
    if starts points are equal, compare the end points
    *this function is named compareInterval and not compareValue since there
    is already a compareValue function in the line 
    """
    def compareInterval(self, other):
        c = self.unitDir.compareValue(other.unitDir) 
        if c != 0:
            return c
        c = self.root.compareValue(other.root) 
        if c != 0:
            return c
        c = self.start.compareValue(other.start) 
        if c != 0:
            return c
        c = self.end.compareValue(other.end) 
        if c != 0:
            return c
        return 0
        
    """
    check if the line segment contains a point
    """
    def containsPoint(self,point):
        if not point.colinear(self.start,self.end):
            return False
        if self.start.compareValue(point) < 1 and self.end.compareValue(point) > -1:
            return True
        elif self.start.compareValue(point) > -1 and self.end.compareValue(point) < 1:
            return True
        return False
        
    def compareEnd(self, other):
        return self.end.compareValue(other.end) 
            
    """
    find the length of the line segment
    """
    def length(self):
        lengthVector = self.end.subtract(self.start)
        return lengthVector.dotproduct(lengthVector) ** coordinate(0.5)
        
    """
    check if two line segments touch one another
             o-----------o  
    o--------o
    
    or
        o-----o
              o------o
    """
    def touches(self,other):
        return (self.end == other.start or self.start == other.end) and self.compareValue(other) == 0 
    
    """
    check if this line segment contains the other line segment
        o---------------------o
                o--------o
    or
        o---------------------o
        o---------o
    
    or
        o----------------------o
                o--------------o
    """
    def contains(self,other):
        return (self.end.compareValue(other.end) >= 0 and self.start.compareValue(other.start) <= 0) and self.compareValue(other) == 0
    
    """
    check if two line segments are disjoint
    o------------o
                        o----------o
    """
    def disjoint(self,other):
        return (self.end.compareValue(other.start) == -1 or self.start.compareValue(other.end) == 1) and self.compareValue(other) == 0
    
    """
    check if two line segments are align
        o------------------o
            o--------------o    
    or
        o------------------o
        o---------o    
    or
                o----------o
        o------------------o 
    or 
        o-----o
        o------------------o
    """
    def align(self, other):
        return (self.end == other.end or self.start == other.start) and self.compareValue(other) == 0
        
    """
    find the common of two line segment:
    o--------------o
            o-----------o
    -->     o------o        
    """
    def common(self,other):
        if self.disjoint(other) or self.touches(other):
            return False
            
        if self.end.compareValue(other.end) == 1:
            self.end = vector2D(other.end.x, other.end.y)
        if self.start.compareValue(other.start) == -1:
            self.start = vector2D(other.start.x, other.start.y)
        return True
        
    """
    combine two line segments into one (two lines cannot be disjoint)
    e.g.
            o----------o
                       o-------o      
       ---> o------------------o
    """
    def combine(self,other):
        if self.disjoint(other):
            return False
        if self.end.compareValue(other.end) == -1:
            self.end = vector2D(other.end.x, other.end.y)
        if self.start.compareValue(other.start) == 1:
            self.start = vector2D(other.start.x, other.start.y)
            
        self.direction = (self.end.subtract(self.start))
        self.unitDir = self.direction.unit()
        self.root = self.start.subtract(self.direction.scale(self.start))
        return True
        
    """
    find the complement of two lines
    e.g.
            o-------------o
                  o------------o      
       ---> o-----o       o----o
    or
            o------------------o
                    o----------o      
       ---> o-------o  
    or
            o------------------o
            o----------o      
       --->            o-------o  
    """
    def complement(self,other):
        result = []
        if self.start.compareValue(other.start) == -1 and self.end.compareValue(other.start) == 1 :
            new = self.duplicateInd(self.sort)
            new.end = other.start
            result.append(new)
        if self.start.compareValue(other.end) == -1 and self.end.compareValue(other.end) == 1 :
            new = self.duplicateInd(self.sort)
            new.start = other.end
            result.append(new)
        if self.hasAttribute():
            for i in result:
                i.addAttribute(self.attribute)
        return result
        
    """
    check if the two line segments intersect one another
    """
    def intersect(self, other):
        if other is None or other.type != 'ind-lineSegment2D' :
            return False
        return not self.direction.parallel(other.direction)
          
    def duplicateInd(self, sort = 0):
        if sort == 0:
            sort = self.sort
        lineSeg =  lineSegment2D(sort, self.start, self.end)
        if self.hasAttribute():
            lineSeg.addAttribute(self.attribute)
        return lineSeg    
        
    """
    return the line that contains this segment
    """
    def codescriptor(self):
        line = line2D(self.sort, self.start, self.end)
        line.assoc = self
        return line
        

    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform2D)
        if affine is None:
            return self.duplicateInd()
        start = self.start.transform(affine)
        end = self.end.transform(affine)         
        return lineSegment2D(self.sort, start, end)
                
    def convertFloat(self):
        a = lineSegment2D(self.sort, self.start.convertFloat(), self.end.convertFloat())
        if self.hasAttribute():
            a.addAttribute(self.attribute.convertFloat())
        return a
        
    def findFloatValue(self, element):
        if element.type != "ind-lineSegment2D":
            if element.type == "ind-line2D":
                if self.codescriptor().findFloatValue(element) is not False:
                    return self
                else:
                    return False
        if (self.start.vectorFloat == element.start and
            self.end.vectorFloat == element.end):
            return self
        if self.hasAttribute():
            return self.attribute.findFloatValue(element)
        return False
        
    def checkTransform(self, transform, compareForm):
        affine = transform.retrieveTransform(affineTransform2D)
        if affine is None:
            tfLineSeg = self
        else:
            start = self.start.transform(affine)
            end = self.end.transform(affine)  
            tfLineSeg = lineSegment2D(self.sort, start, end)
            if self.hasAttribute():
                tfLineSeg.addAttribute(self.attribute)
        if compareForm.containsInd(tfLineSeg):
            return True
        return False
        
    def containsInd(self, ind):
        if not self.contains(ind):
            return False
        if self.hasAttribute():
            if ind.hasAttribute():
                if ind.attribute.partOf(self.attribute):
                    return True
                else:
                    return False
            else:
                return True
        return True


class lineSegment3D(line3D):
    """
    initialize a line segment
    initialize a line containing it first
    find the lesser and greater points between the 2 points passed in (start 
    and end) and store them as self.start and self.end respectively
    """
    def __init__(self, sort, start = vector3D(0,0,0), end = vector3D(0,0,0)):
        super().__init__(sort, start, end)     
        self.type = 'ind-lineSegment3D'
        if end.type == 'ind-point3D':
            end = end.value
        if start.type == 'ind-point3D':
            start = start.value
        if end.compareValue(start) == -1:
            self.start = end
            self.end = start
        elif end.compareValue(start) == 1:
            self.end = end
            self.start = start
        else:
            self.end = end
            self.start = start

    def __str__(self):
        return self.type+'_'+str(self.start)+','+str(self.end)

    def printVal(self):
        return str(self.start)+ '|'+ str(self.end) # + ', direction: ('+str(self.direction.x)+','+str(self.direction.y)+ '), unitDir: ('+str(self.unitDir.x)+','+str(self.unitDir.y)+ '), root: ('+str(self.root.x)+','+str(self.root.y)+ ')'

    def __eq__(self, other):
        if other is None or other.type != 'ind-lineSegment3D' :
            return False
        if self.sort == other.sort:
            return (self.start == other.start) and (self.end == other.end)
        
    def isNil(self):
        if self.end is None or self.start is None:
            return True
        return self.end == self.start
        
    """
    compare the value of two line segments
    first compare using unit directions
    if unit directions are equal, compare using roots
    if roots are equal, compare the start points
    if starts points are equal, compare the end points
    *this function is named compareInterval and not compareValue since there
    is already a compareValue function in the line 
    """
    def compareInterval(self, other):
        c = self.unitDir.compareValue(other.unitDir) 
        if c != 0:
            return c
        c = self.root.compareValue(other.root) 
        if c != 0:
            return c
        c = self.start.compareValue(other.start) 
        if c != 0:
            return c
        c = self.end.compareValue(other.end) 
        if c != 0:
            return c
        return 0
        
    """
    check if the line segment contains a point
    """
    def containsPoint(self,point):
        if not point.colinear(self.start,self.end):
            return False
        if self.start.compareValue(point) < 1 and self.end.compareValue(point) > -1:
            return True
        elif self.start.compareValue(point) > -1 and self.end.compareValue(point) < 1:
            return True
        return False
        
    def compareEnd(self, other):
        return self.end.compareValue(other.end) 
            
    """
    find the length of the line segment
    """
    def length(self):
        lengthVector = self.end.subtract(self.start)
        return lengthVector.dotproduct(lengthVector) ** coordinate(0.5)
        
    """
    check if two line segments touch one another
             o-----------o  
    o--------o
    
    or
        o-----o
              o------o
    """
    def touches(self,other):
        return (self.end == other.start or self.start == other.end) and self.compareValue(other) == 0 
    
    """
    check if this line segment contains the other line segment
        o---------------------o
                o--------o
    or
        o---------------------o
        o---------o
    
    or
        o----------------------o
                o--------------o
    """
    def contains(self,other):
        return (self.end.compareValue(other.end) >= 0 and self.start.compareValue(other.start) <= 0) and self.compareValue(other) == 0
    
    """
    check if two line segments are disjoint
    o------------o
                        o----------o
    """
    def disjoint(self,other):
        return (self.end.compareValue(other.start) == -1 or self.start.compareValue(other.end) == 1) and self.compareValue(other) == 0
    
    """
    check if two line segments are align
        o------------------o
            o--------------o    
    or
        o------------------o
        o---------o    
    or
                o----------o
        o------------------o 
    or 
        o-----o
        o------------------o
    """
    def align(self, other):
        return (self.end == other.end or self.start == other.start) and self.compareValue(other) == 0
        
    """
    find the common of two line segment:
    o--------------o
            o-----------o
    -->     o------o        
    """
    def common(self,other):
        if self.disjoint(other) or self.touches(other):
            return False
            
        if self.end.compareValue(other.end) == 1:
            self.end = other.end
        if self.start.compareValue(other.start) == -1:
            self.start = other.start
        return True
        
    """
    combine two line segments into one (two lines cannot be disjoint)
    e.g.
            o----------o
                       o-------o      
       ---> o------------------o
    """
    def combine(self,other):
        if self.disjoint(other):
            return False
        if self.end.compareValue(other.end) == -1:
            self.end = other.end
        if self.start.compareValue(other.start) == 1:
            self.start = other.start
            
        self.direction = (self.end.subtract(self.start))
        self.unitDir = self.direction.unit()
        self.root = self.start.subtract(self.direction.scale(self.start))
        return True
        
    """
    find the complement of two lines
    e.g.
            o-------------o
                  o------------o      
       ---> o-----o       o----o
    or
            o------------------o
                    o----------o      
       ---> o-------o  
    or
            o------------------o
            o----------o      
       --->            o-------o  
    """
    def complement(self,other):
        result = []
        if self.start.compareValue(other.start) == -1 and self.end.compareValue(other.start) == 1 :
            new = self.duplicateInd(self.sort)
            new.end = other.start
            result.append(new)
        if self.start.compareValue(other.end) == -1 and self.end.compareValue(other.end) == 1 :
            new = self.duplicateInd(self.sort)
            new.start = other.end
            result.append(new)
        if self.hasAttribute():
            for i in result:
                i.addAttribute(self.attribute)
        return result
                  
    def duplicateInd(self, sort = 0):
        if sort == 0:
            sort = self.sort
        lineSeg =  lineSegment3D(sort, self.start, self.end)
        if self.hasAttribute():
            lineSeg.addAttribute(self.attribute)
        return lineSeg    
        
    """
    return the line that contains this segment
    """
    def codescriptor(self):
        line = line3D(self.sort, self.start, self.end)
        line.assoc = self
        return line
        

    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            return self.duplicateInd()
        start = self.start.transform(affine)
        end = self.end.transform(affine)         
        return lineSegment3D(self.sort, start, end)
                
    def convertFloat(self):
        a = lineSegment3D(self.sort, self.start.convertFloat(), self.end.convertFloat())
        if self.hasAttribute():
            a.addAttribute(self.attribute.convertFloat())
        return a
        
    def findFloatValue(self, element):
        if element.type != "ind-lineSegment3D":
            if element.type == "ind-line3D":
                if self.codescriptor().findFloatValue(element) is not False:
                    return self
                else:
                    return False
        if (self.start.vectorFloat == element.start and
            self.end.vectorFloat == element.end):
            return self
        if self.hasAttribute():
            return self.attribute.findFloatValue(element)
        return False
        
    def checkTransform(self, transform, compareForm):
        affine = transform.retrieveTransform(affineTransform3D)
        if affine is None:
            tfLineSeg = self
        else:
            start = self.start.transform(affine)
            end = self.end.transform(affine)  
            tfLineSeg = lineSegment3D(self.sort, start, end)
            if self.hasAttribute():
                tfLineSeg.addAttribute(self.attribute)
        if compareForm.containsInd(tfLineSeg):
            return True
        return False
        
    def containsInd(self, ind):
        if not self.contains(ind):
            return False
        if self.hasAttribute():
            if ind.hasAttribute():
                if ind.attribute.partOf(self.attribute):
                    return True
                else:
                    return False
            else:
                return True
        return True
