# -*- coding: utf-8 -*-
"""
Created on Mon Nov  7 10:41:33 2016

@author: akibdpt
"""
from sympy import sqrt
from ..struct.coordinate import coordinate
from ..struct.vector import vector2D, vector3D

class bSegment():
    POS_X_INF = vector2D(1,0,True)
    POS_Y_INF = vector2D(0,1,True)
    NEG_Y_INF = vector2D(0,-1,True)
    
    POS_X_INF_3D = vector3D(1,0,0,True)
    POS_Y_INF_3D = vector3D(0,1,0,True)
    NEG_Y_INF_3D = vector3D(0,-1,0,True)
    
    INFINITE = -1
    CLOSED = 0
    OPEN = 1
    
    def __init__(self, start, end, assoc = None):
        self.type = 'struct-bSegment'
        if 'ind' in start.type or 'ind' in end.type:
            raise ValueError('start and end must be vectors')
        if start.compareValue(end) == 1:
            self.start = end
            self.end = start
        else:
            self.start = start
            self.end = end
        self.assoc = assoc
        self.alignedPositive = False
        
    def __str__(self):
        return 'bs'+str(self.start)+str(self.end)
        
    def direction(self):
        return self.end.subtract(self.start)
        
    def root(self):
        return self.start.subtract(self.direction().scale(self.start))
    
    def length(self):
        direction = self.end.subtract(self.start)
        return direction.dotproduct(direction) ** coordinate(0.5)
        
    def squareLength(self):
        direction = self.end.subtract(self.start)
        return direction.dotproduct(direction)
        
    def isInfinite(self):
        return self.start.isInfinite() or self.end.isInfinite()
    
    def aligns(self, other):
        return (self.end.colinear(other.start, other.end) and 
                self.direction().unit() == other.direction().unit())
                
    def partOf(self, other):
        return (self.end.colinear(other.start, other.end) and 
                self.direction().unit() == other.direction().unit() and
                self.start >= other.start and self.end <= other.end)
    
    def __eq__(self, other):
        if other is None or other.type != 'struct-bSegment' :
            return False
        return (self.start == other.start) and (self.end == other.end)
            
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1
        
    def __le__(self, other):
        return self.compareValue(other) < 1
        
    def __ge__(self, other):
        return self.compareValue(other) > -1
    
    def compareValue(self, other):
        c = self.direction().unit().compareValue(other.direction().unit()) 
        if c != 0:
            return c
        c = self.root().compareValue(other.root()) 
        if c != 0:
            return c
        c = self.start.compareValue(other.start) 
        if c != 0:
            return c
        return self.end.compareValue(other.end) 
        
    def transform(self, matrix):
        return bSegment(self.start.transform(matrix), self.end.transform(matrix))
    
    def intersection(self, other, type = CLOSED):
        if self.isInfinite() or other.isInfinite():
            return None
            
        dir1 = self.direction()
        dir2 = other.direction()
        
        if dir1.parallel(dir2): 
            return None

        cross = dir1.product(dir2)
        diff = other.start.subtract(self.start)
        
        if self.start.type == 'struct-vector2D' or self.start.type == 'struct-vector2Dfloat':
            par1 = diff.product(dir2) / cross
            par2 = diff.product(dir1) / cross
        else:
            denom = cross.dotproduct(cross)
            par1 = diff.product(dir2).dotproduct(cross) / denom
            par2 = diff.product(dir1).dotproduct(cross) / denom
            
        if (par2>=0 and par2<=1 and (type == bSegment.INFINITE or 
            (type == bSegment.CLOSED and par1 >= 0 and par1 <= 1) or
            (type == bSegment.OPEN and par1 > 0 and par1< 1))):
            return self.start.add(dir1.scale(par1)).simplify()
        return None

