# -*- coding: utf-8 -*-
"""
Created on Fri Aug 12 14:33:25 2016

@author: akibdpt
"""
from .coordinate import coordinate
from .vectorFloat import vector2Dfloat, vector3Dfloat
from sympy import acos,  pi, N
from math import atan2
#from decimal import Decimal
from mpmath import mp

mp.dps = vector2Dfloat.mpdps

"""
 * A vector specifies a position in a two-dimensional cartesian space.
 * If normalized, it only specifies a direction.
 * 
 * The Vector class defines a vector as a pair of {coordinate}'s
 * and a w factor to reflect the vector's infinity characteristic
 * A Vector object is never modified after creation, thus,
 * it can be used multiple times.
"""
class vector2D():
    """
    initialize vector
    containing x coordinate, y coordinate and w
    w is indicator of infinity. If w is 1 -> this is not an infinity vector
                                If w is 0 -> this is an infinity vector with direction(x,y)
    """
    
    def __init__(self, x, y, infinity = False):
        self.type = 'struct-vector2D'
        if infinity:
            self.w = 0
        else:
            self.w = 1
        if isinstance(x, coordinate):
            self.x = coordinate(x.value)
        else:
            self.x = coordinate(x)
        if isinstance(y, coordinate):
            self.y = coordinate(y.value)
        else:
            self.y = coordinate(y)
        self.floatVector = None

    """
    getter for x value
    """
    @property
    def x(self):
        if self.w != 0:
            return self.__x / self.w
        else:
            return self.__x
        
    """
    setter for x value
    """
    @x.setter
    def x(self, val):
        if self.w != 0:
            self.__x = val * self.w
        else:
            self.__x = val
            
    """
    getter for y value
    """
    @property
    def y(self):
        if self.w != 0:
            return self.__y / self.w
        else:
            return self.__y
            
    """
    setter for y value
    """
    @y.setter
    def y(self, val):
        if self.w != 0:
            self.__y = val * self.w
        else:
            self.__y = val
     
    """
    return the string that represents this vector's values
    """       
    def __str__(self):
        if self.w == 0:
            return ' inf('+str(self.x)+','+str(self.y)+')'
        return '('+str(self.x)+','+str(self.y)+')'
        
        
    def __abs__(self):
        if self.sign() >= 0:
            return self
        else:
            return vector2D(-self.x, -self.y)
        
    def printVal(self):
        if self.w == 0:
            return 'inf('+self.x+','+self.y+')'
        return '('+self.x+','+self.y+')'

    """
    return the string that represents this vector's values in floating point
    """
    def evalf(self):
        if self.w == 0:
            return 'inf('+self.x.evalf()+','+self.y.evalf()+')'
        return '('+self.x.evalf()+','+self.y.evalf()+')'
    
    """
    simplifying the expressions in the vector's values
    e.g. vector.x = sin(2*pi/3) will be simplified into sqrt(3)/2
    * this function is rather inefficient in running time, 
    * try to minimize the use of this function 
    """
    def simplify(self):
        self.x = self.x.simplify()
        self.y = self.y.simplify()
        return self

    """
    converting the vector into a vector that contains only float values
    """
    def convertFloat(self):
        """
        return vector2Dfloat(mp.mpf(N(self.x.value,156)), mp.mpf(N(self.y.value,156)), self.w)
        """
        if self.floatVector is None:
            self.floatVector = vector2Dfloat(mp.mpf(N(self.x.value,vector2Dfloat.mpdps+1)), mp.mpf(N(self.y.value,vector2Dfloat.mpdps+1)), self.w)
        return self.floatVector

    """
    find the sign of the vector
    the vector is positive if the x value is positive
    the vector is negative if the x value is negative
    if x value is 0, the sign of the vector is decided by the y value
    return 1 if positive, -1 if negative and 0 if both x and y are 0
    """
    def sign(self):
        if self.x > 0:
            return 1
        if self.x < 0:
            return -1
        if self.y > 0:
            return 1
        if self.y < 0:
            return -1
        return 0
    
    def isZero(self):
        return self.x == 0 and self.y == 0
        
    """
    checking if two vectors are parallel
    two vectors are parallel if: this.x / other.x = this.y / other.y
                             --> this.x * other.y = this.y * other.x
    """
    def parallel(self,other):
        if other.type != 'struct-vector2D':
            return False
        if self.isZero() or other.isZero():
            return True 
        lhs = (self.x * other.y).evalf()
        rhs = (self.y * other.x).evalf()
        return lhs == rhs
        
    """
    checking if two vectors are perpendicular with one another
    two vectors are perpendicular if their dot product value is 0
    """
    def perpendicular(self,other):
        if other.type != 'struct-vector2D':
            return False
        return self.dotproduct(other) == 0
        
    def isInfinite(self):
        return self.w == 0
        
    """
    checking if this vector is equal to other vector
    two vectors are equal if both have equal x values and equal y values
    """
    def __eq__(self, other):
        if other.type != 'struct-vector2D':
            return False
        return self.x == other.x  and self.y  == other.y and self.w == other.w
        
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __le__(self, other):
        return self.compareValue(other) < 1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1
        
    def __ge__(self, other):
        return self.compareValue(other) > -1

    """
    comparing the value of this vector with another vector
    a vector is greater if its x value is greater and vice versa
    if the x values of the two vectors are the same, the vector with the 
    greater y value is greater
    
    return 1 if this vector is greater than the other vector
    return -1 if this vector is lesser than the other vector
    return 0 if both vectors have equal x and y values
    """
    def compareValue(self, other):
        if self.w == other.w:
            if self.x > other.x:
                return 1
            elif self.x < other.x:
                return -1
            elif self.y > other.y:
                return 1
            elif self.y < other.y:
                return -1
            return 0
        elif self.w == 0:
            if self.x < 0:
                return -1
            elif self.x > 0:
                return 1
            elif other.x < 0:
                return 1
            elif other.x > 0:
                return -1
            else:
                if self.y < 0:
                    return -1
                elif self.y > 0:
                    return 1
                elif other.y < 0:
                    return 1
                elif other.y > 0:
                    return -1
                else:
                    return 0
        elif other.w == 0:
            if other.x < 0:
                return 1
            elif other.x > 0:
                return -1
            elif self.x < 0:
                return -1
            elif self.x > 0:
                return 1
            else:
                if other.y < 0:
                    return 1
                elif other.y > 0:
                    return -1
                elif self.y < 0:
                    return -1
                elif self.y > 0:
                    return 1
                else:
                    return 0
    
    """
    scaling the vector by a certain factor
    if factor is a number: multiply the vector's x and y with the factor
    if factor is another vector: multiply the vector's x and y with the projection length of the other vector
    """    
    def scale(self, other):
        if isinstance(other, coordinate):
            return vector2D(self.x*other, self.y*other)
        else:
            scalar = (other.dotproduct(self)) / (self.dotproduct(self))
            return vector2D(self.x*scalar, self.y*scalar)
            
    """
    find the vector with absolute values of this vector's x and y
    """
    def abs(self):
        return self.scale(self.sign())
                
    def inverse(self):
        return vector2D(-self.x, -self.y)
        
    """
    find the length of the vector
    """
    def length(self):
        return (self.x * self.x + self.y *self.y) ** coordinate(1/2)
        
    """
    find the squared length of the vector
    *if length equality is to be checked, use squared length since it is more 
     time efficient (no need to calculate square root)
    """
    def squareLength(self):
        return (self.x * self.x + self.y *self.y)
        
    """
    find the unit vector of the vector: dividing the vector's x and y by the length of the vector
    """
    def unit(self):
        a = coordinate(1) / self.length()
        return vector2D(self.x * a, self.y * a)
    
    """
    adding two vector
    """
    def add(self,other):
        if self.w == 0:
            if other.w == 0:
                raise ValueError('Adding two infinite vectors is undefined')
            else:
                return self
        if other.w == 0:
            return other
        return vector2D(self.x+other.x, self.y+other.y)
    
    """
    subtracting the vector by another vector
    """
    def subtract(self, other):
        if self.w == 0:
            if other.w == 0 and self != other.inverse():
                return vector2D(self.x-other.x, self.y-other.y, True)
            else:
                return self
        if other.w == 0:
            return other
        return vector2D(self.x-other.x, self.y-other.y)
        
    """
    find the dot product of the vector with another
    """
    def dotproduct(self, other):
        return self.x * other.x + self.y * other.y
        
    """
    find the product of the vector with another,
    since the vectors are in 2D space, the vectors will be (x0,y0,0) and (x1,y1,0)
    in 3D space --> the product vector will be (0, 0, x0*y1-x1*y0)
    this function return only the z value of the product vector since the vector
    itself cannot be represented in the 2D space
    """
    def product(self, other):
        return self.x * other.y - self.y * other.x
    
    """    
    """
    def scalar(self, other):
        return other.dotproduct(other) / self.dotproduct(self)
    
    """
    check if this vector and 2 other vectors are colinear
    """
    def colinear(self, other1, other2):
        return other1.subtract(self).parallel(other2.subtract(self))
    
    """
    find the clockwise angle from this vector to another vector
    """
    def angle(self, other):
        a = self.unit()
        b = other.unit()
        #return (atan2(b.y, b.x) - atan2(a.y, a.x))
        
        c = atan2(b.y.evalf(), b.x.evalf()) - atan2(a.y.evalf(), a.x.evalf())
        if (c < 0 and c > -pi) or (c > pi and c < 2*pi):            
            return - acos(self.unit().dotproduct(other.unit()).value)
        return acos(self.unit().dotproduct(other.unit()).value)
                
    """
    transforming the vector with a transformation matrix
    matrix: [[a, b, c]
             [d, e, f]]
    this vector: [[x]
                  [y]
                  [1]]
    => vector after transforming: [[a*x + b*y + c]
                                   [d*x + e*y + f]
                                   [  who cares  ]]
    """
    def transform(self, transf):
        return vector2D(((self.x*transf.matrix[0][0] + self.y*transf.matrix[0][1] + transf.matrix[0][2])).simplify(),
                        ((self.x*transf.matrix[1][0] + self.y*transf.matrix[1][1] + transf.matrix[1][2])).simplify())

    
"""

"""
class vector3D():
    def __init__(self, x, y, z, infinity = False):
        self.type = 'struct-vector3D'
        if infinity:
            self.w = coordinate(0)
        else:
            self.w = coordinate(1)
        if isinstance(x, int) or isinstance(x, float):
            self.x = coordinate(x)
        else:
            self.x = coordinate(x.value)
        if isinstance(y, int) or isinstance(y, float):
            self.y = coordinate(y)
        else:
            self.y = coordinate(y.value)
        if isinstance(z, int) or isinstance(z, float):
            self.z = coordinate(z)
        else:
            self.z = coordinate(z.value)
            
    @property
    def x(self):
        if self.w != 0:
            return self.__x / self.w
        else:
            return self.__x
        
    @x.setter
    def x(self, val):
        if self.w != 0:
            self.__x = val * self.w
        else:
            self.__x = val
            
    @property
    def y(self):
        if self.w != 0:
            return self.__y / self.w
        else:
            return self.__y
        
    @y.setter
    def y(self, val):
        if self.w != 0:
            self.__y = val * self.w
        else:
            self.__y = val
            
    @property
    def z(self):
        if self.w != 0:
            return self.__z / self.w
        else:
            return self.__z
        
    @z.setter
    def z(self, val):
        if self.w != 0:
            self.__z = val * self.w
        else:
            self.__z = val
            
    """
    return the string that represents this vector's values
    """       
    def __str__(self):
        if self.w == 0:
            return ' inf('+str(self.x)+','+str(self.y)+','+str(self.z)+')'
        return '('+str(self.x)+','+str(self.y)+','+str(self.z)+')'
        
    def __abs__(self):
        if self.sign() >= 0:
            return self
        else:
            return vector3D(-self.x, -self.y, -self.z)
        
    def printVal(self):
        if self.w == 0:
            return 'inf('+self.x+','+self.y+','+self.z+')'
        return '('+self.x+','+self.y+','+self.z+')'

    """
    return the string that represents this vector's values in floating point
    """
    def evalf(self):
        if self.w == 0:
            return 'inf('+self.x.evalf()+','+self.y.evalf()+','+self.z.evalf()+')'
        return '('+self.x.evalf()+','+self.y.evalf()+','+self.z.evalf()+')'
    
    """
    simplifying the expressions in the vector's values
    e.g. vector.x = sin(2*pi/3) will be simplified into sqrt(3)/2
    * this function is rather inefficient in running time, 
    * try to minimize the use of this function 
    """
    def simplify(self):
        return vector3D(self.x.simplify(), self.y.simplify(), self.z.simplify())

    """
    converting the vector into a vector that contains only float values
    """
    def convertFloat(self):
        try:
            return vector3Dfloat(mp.mpf(self.x.value), mp.mpf(self.y.value),
                                 mp.mpf(self.z.value), self.w)
        except TypeError:
            return vector3Dfloat(mp.mpf(str(N(self.x.value,vector2Dfloat.mpdps))), 
                                 mp.mpf(str(N(self.y.value,vector2Dfloat.mpdps))),
                                 mp.mpf(str(N(self.z.value,vector2Dfloat.mpdps))), 
                                 self.w)


    """
    find the sign of the vector
    the vector is positive if the x value is positive
    the vector is negative if the x value is negative
    if x value is 0, the sign of the vector is decided by the y value
    return 1 if positive, -1 if negative and 0 if both x and y are 0
    """
    def sign(self):
        if self.x > 0:
            return 1
        if self.x < 0:
            return -1
        if self.y > 0:
            return 1
        if self.y < 0:
            return -1
        if self.z > 0:
            return 1
        if self.z < 0:
            return -1
        return 0
    
    def isZero(self):
        return self.x == 0 and self.y == 0 and self.z == 0
        
    """
    checking if two vectors are parallel
    two vectors are parallel if: this.x / other.x = this.y / other.y
                             --> this.x * other.y = this.y * other.x
    """
    def parallel(self,other):
        if other.type != 'struct-vector3D':
            return False
        if self.isZero() or other.isZero():
            return True 
        if self.x != 0 or other.x != 0:
            lhs1 = (self.x * other.y).evalf()
            rhs1 = (self.y * other.x).evalf()
            
            lhs2 = (self.x * other.z).evalf()
            rhs2 = (self.z * other.x).evalf()
            
            return lhs1 == rhs1 and lhs2 == rhs2
        lhs = (self.y * other.z).evalf()
        rhs = (self.z * other.y).evalf()
        return lhs == rhs
        
    """
    checking if two vectors are perpendicular with one another
    two vectors are perpendicular if their dot product value is 0
    """
    def perpendicular(self,other):
        if other.type != 'struct-vector3D':
            return False
        return self.dotproduct(other) == 0
        
    def isInfinite(self):
        return self.w == 0
        
    """
    checking if this vector is equal to other vector
    two vectors are equal if both have equal x values and equal y values
    """
    def __eq__(self, other):
        if other.type != 'struct-vector3D':
            return False
        return (self.x == other.x  and self.y  == other.y and 
                self.z  == other.z and self.w == other.w)
        
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __le__(self, other):
        return self.compareValue(other) < 1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1
        
    def __ge__(self, other):
        return self.compareValue(other) > -1

    """
    comparing the value of this vector with another vector
    a vector is greater if its x value is greater and vice versa
    if the x values of the two vectors are the same, the vector with the 
    greater y value is greater
    
    return 1 if this vector is greater than the other vector
    return -1 if this vector is lesser than the other vector
    return 0 if both vectors have equal x and y values
    """
    def compareValue(self, other):
        if self.w == other.w:
            if self.x > other.x:
                return 1
            elif self.x < other.x:
                return -1
            elif self.y > other.y:
                return 1
            elif self.y < other.y:
                return -1
            elif self.z > other.z:
                return 1
            elif self.z < other.z:
                return -1
            return 0
        elif self.w == 0:
            if self.x < 0:
                return -1
            elif self.x > 0:
                return 1
            elif other.x < 0:
                return 1
            elif other.x > 0:
                return -1
            else:
                if self.y < 0:
                    return -1
                elif self.y > 0:
                    return 1
                elif other.y < 0:
                    return 1
                elif other.y > 0:
                    return -1
                else:
                    if self.z < 0:
                        return -1
                    elif self.z > 0:
                        return 1
                    elif other.z < 0:
                        return 1
                    elif other.z > 0:
                        return -1
                    else:
                        return 0
        elif other.w == 0:
            if other.x < 0:
                return 1
            elif other.x > 0:
                return -1
            elif self.x < 0:
                return -1
            elif self.x > 0:
                return 1
            else:
                if other.y < 0:
                    return 1
                elif other.y > 0:
                    return -1
                elif self.y < 0:
                    return -1
                elif self.y > 0:
                    return 1
                else:
                    if other.z < 0:
                        return 1
                    elif other.z > 0:
                        return -1
                    elif self.z < 0:
                        return -1
                    elif self.z > 0:
                        return 1
                    else:
                        return 0
    
    """
    scaling the vector by a certain factor
    if factor is a number: multiply the vector's x and y with the factor
    if factor is another vector: multiply the vector's x and y with the projection length of the other vector
    """    
    def scale(self, other):
        if isinstance(other, coordinate):
            return vector3D(self.x*other, self.y*other, self.z*other)
        else:
            scalar = (other.dotproduct(self)) / (self.dotproduct(self))
            return vector3D(self.x*scalar, self.y*scalar, self.z*scalar)
            
    """
    find the vector with absolute values of this vector's x and y
    """
    def abs(self):
        return self.scale(self.sign())
                
    def inverse(self):
        return vector3D(-self.x, -self.y, -self.z)
        
    """
    find the length of the vector
    """
    def length(self):
        return (self.x * self.x + self.y *self.y + self.z *self.z) ** coordinate(1/2)
        
    """
    find the squared length of the vector
    *if length equality is to be checked, use squared length since it is more 
     time efficient (no need to calculate square root)
    """
    def squareLength(self):
        return (self.x * self.x + self.y *self.y + self.z *self.z)
        
    """
    find the unit vector of the vector: dividing the vector's x and y by the length of the vector
    """
    def unit(self):
        a = coordinate(1) / self.length()
        return vector3D(self.x * a, self.y * a, self.z * a)
    
    """
    adding two vector
    """
    def add(self,other):
        if self.w == 0:
            if other.w == 0:
                raise ValueError('Adding two infinite vectors is undefined')
            else:
                return self
        if other.w == 0:
            return other
        return vector3D(self.x+other.x, self.y+other.y, self.z+other.z)
    
    """
    subtracting the vector by another vector
    """
    def subtract(self, other):
        if self.w == 0:
            if other.w == 0 and self != other.inverse():
                return vector3D(self.x-other.x, self.y-other.y, self.z-other.z, True)
            else:
                return self
        if other.w == 0:
            return other
        return vector3D(self.x-other.x, self.y-other.y, self.z-other.z)
        
    """
    find the dot product of the vector with another
    """
    def dotproduct(self, other):
        return self.x * other.x + self.y * other.y + self.z * other.z
        
    """
    find the product of the vector with another,
    since the vectors are in 2D space, the vectors will be (x0,y0,0) and (x1,y1,0)
    in 3D space --> the product vector will be (0, 0, x0*y1-x1*y0)
    this function return only the z value of the product vector since the vector
    itself cannot be represented in the 2D space
    """
    def product(self, other):
        return vector3D(self.y * other.z - self.z * other.y,
                        self.z * other.x - self.x * other.z,
                        self.x * other.y - self.y * other.x)
    
    """    
    """
    def scalar(self, other):
        return other.dotproduct(other) / self.dotproduct(self)
    
    """
    check if this vector and 2 other vectors are colinear
    """
    def colinear(self, other1, other2):
        return other1.subtract(self).parallel(other2.subtract(self))
    
    
    ###########################################################################
    """
    def angle(self, other):
        a = self.unit()
        b = other.unit()
        #return (atan2(b.y, b.x) - atan2(a.y, a.x))
        
        c = atan2(b.y.evalf(), b.x.evalf()) - atan2(a.y.evalf(), a.x.evalf())
        if (c < 0 and c > -pi) or (c > pi and c < 2*pi):            
            return - acos(self.unit().dotproduct(other.unit()).value)
        return acos(self.unit().dotproduct(other.unit()).value)
    """
    ###########################################################################

    """
    transforming the vector with a transformation matrix
    matrix: [[a, b, c]
             [d, e, f]]
    this vector: [[x]
                  [y]
                  [1]]
    => vector after transforming: [[a*x + b*y + c]
                                   [d*x + e*y + f]
                                   [  who cares  ]]
    """
    def transform(self, transf):
        return vector3D((self.x*transf.matrix[0][0] + self.y*transf.matrix[0][1] + self.z*transf.matrix[0][2] + transf.matrix[0][3]).simplify(),
                        (self.x*transf.matrix[1][0] + self.y*transf.matrix[1][1] + self.z*transf.matrix[1][2] + transf.matrix[1][3]).simplify(),
                        (self.x*transf.matrix[2][0] + self.y*transf.matrix[2][1] + self.z*transf.matrix[2][2] + transf.matrix[2][3]).simplify())
            